package cytoPrefuse;

import java.awt.event.InputEvent;
import java.awt.event.MouseEvent;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.Iterator;

import javax.swing.JFileChooser;

import prefuse.Display;
import prefuse.Visualization;
import prefuse.controls.ControlAdapter;
import prefuse.data.Edge;
import prefuse.data.Node;
import prefuse.data.expression.Predicate;
import prefuse.visual.NodeItem;

public class SaveAction extends ControlAdapter {
	  public void mouseClicked(MouseEvent e) {
		  if ((e.getModifiers() & InputEvent.BUTTON3_MASK)==InputEvent.BUTTON3_MASK  &&
				  (e.getModifiers() & InputEvent.SHIFT_MASK)==InputEvent.SHIFT_MASK) {
		        Display display = (Display)e.getComponent();
	            Visualization vis = display.getVisualization();
	           
	            String filename;
	            final JFileChooser fc	 = new JFileChooser();
	    		int retval = fc.showSaveDialog(null );
	    		if (retval ==JFileChooser.APPROVE_OPTION){
	    			File saveFile = fc.getSelectedFile();
	   				filename=saveFile.getPath();
	   				if (filename.contains( ".gml")){
	   					WriteGML (filename,vis);
	   				}else if (filename.contains(".tlp")){
	   					WriteTulip (filename, vis);
	   				}
	   			}
		  	}
	    }
	  
 	  
    private void WriteTulip (String filename, Visualization vis) {
    	try {
			BufferedWriter out = new BufferedWriter(new FileWriter (filename));
			out.write ("(nodes");
			Iterator nodes = vis.items("graph.nodes");
			while (nodes.hasNext()){					
				NodeItem node = (NodeItem) nodes.next();
				out.write (" "+ node.getRow());
			}
			out.write (")");
			out.newLine();

			Iterator edges = vis.items("graph.edges");
			while (edges.hasNext()) {
				Edge edge = (Edge) edges.next();
				int sourceLevel = edge.getSourceNode().getInt( CytoPrefuse.BIO_CATEGORY);
				int destLevel = edge.getTargetNode().getInt( CytoPrefuse.BIO_CATEGORY);
				if (sourceLevel < destLevel) {
					out.write ("(edge " + edge.getRow() + " " + edge.getSourceNode().getRow() + " " + 
							edge.getTargetNode().getRow() + ")");
				}else {
					out.write ("(edge " + edge.getRow() + " " + edge.getTargetNode().getRow() + " " + 
							edge.getSourceNode().getRow() + ")");
				}
				out.newLine();
			}
			
			nodes = vis.items("graph.nodes");
			out.write ("(property  0 string \"viewLabel\"");
			out.newLine();
			out.write ("(default \"\" \"\" )");
			out.newLine();
			while (nodes.hasNext()){					
				NodeItem node = (NodeItem) nodes.next();
				out.write ("(node "+ node.getRow()+ " \""+ node.getString(CytoPrefuse.LABEL) + "\" )");
				out.newLine();
			}
			out.write (")");
			
		
			out.newLine();
					
			int edgeID = 10000;
			for (int i =0; i < 3; i++) {
		
				Predicate curLevel = (Predicate) prefuse.data.expression.parser.ExpressionParser.parse("[bioCategory]="+i);

				Predicate nextLevel = (Predicate) prefuse.data.expression.parser.ExpressionParser.parse("[bioCategory]="+(i+1));  
				nodes = vis.items("graph.nodes", curLevel);
				while (nodes.hasNext()) {
					Node node = (Node) nodes.next();
					Iterator nextNodes = vis.items("graph.nodes", nextLevel);
					while (nextNodes.hasNext()) {
						Node nextNode = (Node) nextNodes.next();
						out.write ("(edge " + edgeID++  + " " + node.getRow() + " " + 
								nextNode.getRow() + ")");
						out.newLine();			
					}
				}			
    		}
			out.newLine();
			out.write ("(property  0 color \"viewColor\"");
			out.newLine();
			out.write ("(default \"(0,255,0,255)\" \"(0,0,0,255)\" )");

			for (int i=10000; i < edgeID; i++) {
				out.write ("(edge " + i+ " \"(0,0,0,0)\")");
				out.newLine();
			};
			out.write (")");
				

			out.close();
    	} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		
    }

	private void WriteGML (String filename, Visualization vis) {
		BufferedWriter out = WriteHeader(filename);
		SaveNodes(out, vis);
		SaveEdges(out, vis);
		WriteFooter(out);
	}
	private void WriteFooter(BufferedWriter out) {
		try {
			out.write ("]");			
			out.newLine();
			out.close();
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}

	private void SaveEdges(BufferedWriter out, Visualization vis) {
		Iterator edges = vis.items("graph.edges");
		try {
			while (edges.hasNext()) {
				Edge edge = (Edge) edges.next();
				out.write ("\tedge");
				out.newLine();
				out.write ("\t[");
				out.newLine();
				out.write ("\t\tsource " + edge.getSourceNode().getRow() );
				out.newLine();
				out.write ("\t\ttarget " + edge.getTargetNode().getRow());
				out.newLine();
				out.write ("\t\tlabel \"");
				if (edge.getBoolean(CytoPrefuse.DNAINTERACTION)) {
					out.write ("pd");
				}else {
					out.write ("pp");
				}
				out.write ("\"");
				out.newLine();
				out.write ("\t]");
				out.newLine();
			}
		}catch (IOException e ) {
			e.printStackTrace();
		}
	}

	private void SaveNodes(BufferedWriter out, Visualization vis) {
		Iterator nodes = vis.items("graph.nodes");
		try {			
			while (nodes.hasNext()){					
				NodeItem node = (NodeItem) nodes.next();
				out.write ("\tnode");
				out.newLine();
				out.write ("\t[");
				out.newLine();
				out.write ("\t\tid "+ node.getRow());
				out.newLine();
				out.write ("\t\tlabel \"" + node.getString(CytoPrefuse.LABEL) + "\"");
				out.newLine();
				out.write ("\t\tgraphics ");
				out.newLine();
				out.write ("\t\t[");
				out.newLine();
				out.write ("\t\t\tx " + node.getX());
				out.newLine();
				out.write ("\t\t\ty " + node.getY());
				out.newLine();
				out.write("\t\t]");
				out.newLine();
				out.write("\t]");
				out.newLine();
			}
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}

	private BufferedWriter WriteHeader(String filename) {
		BufferedWriter out = null;
		try {
			out = new BufferedWriter(new FileWriter (filename));
			out.write( "Creator \"BioGraph\"");
			out.newLine();
			out.write( "Version 1.0");
			out.newLine();
			out.write("graph");
			out.newLine();
			out.write ("[");
			out.newLine();
			out.write( "\tlabel \"\"");
			out.newLine();
			out.write(" \tdirected 0");
			out.newLine();		
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		return out;
	}
}
