package cytoPrefuse;

import java.awt.Point;
import java.awt.geom.Rectangle2D;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Random;
import java.util.Set;

import prefuse.action.layout.Layout;
import prefuse.data.Edge;
import prefuse.data.Node;
import prefuse.data.expression.BooleanLiteral;
import prefuse.data.expression.NotPredicate;
import prefuse.data.expression.Predicate;
import prefuse.data.tuple.TupleSet;
import prefuse.render.LineDividerRenderer;
import prefuse.util.display.DisplayLib;
import prefuse.visual.NodeItem;
import prefuse.visual.VisualGraph;
import prefuse.visual.VisualItem;

public class SearchGridLayout extends Layout {

        private int gridHeight;
        private int gridWidth;
        
        private int totalImprovements;  
        int iteration;
        private List[] candidatePoints;
        private Node[][] nodeAtPosition;
        private List nodesToSample;
        private int [] nodeCounts;
        private int totalNodesFiltered;
        private int[] layerBounds;
        
        private List nodeSampling;
        private List pointSampling;
        
        private SpaceGrid spaceGrid;
        
        public static int SCALE = 60;
        private long  iterationTime;
        private String cancelAction;
        private int iterationsWithoutImprovement;
        private Predicate filter;
        private int leftBorder;
        private int width;
        private String animateAction;
        private Point neighbPoint = new Point();
        private Random generator = new Random( );
        private String categoryKey;        
        private Rectangle2D zoomBounds = new Rectangle2D.Float();
        private Point[] nodepos;
        private Point nodeOrigSave = new Point();
        private boolean firstRun;       
        private VisualGraph graph;
        private int numLayers;
        private LineDividerRenderer lineDividerRenderer;
        public  SearchGridLayout(VisualGraph inGraph, String inCancelAction, String inAnimateAction, int numLayers, LineDividerRenderer lineDividerRenderer) {
                
                 
            cancelAction = inCancelAction;
        
            graph = inGraph;
            nodepos = new Point[graph.getNodeCount()+1];           
            filter = new BooleanLiteral(true);            
            width =(int) (2*(Math.sqrt(graph.getNodeCount())));                    
            leftBorder = 0;
            animateAction = inAnimateAction;                
            categoryKey = CytoPrefuse.BIO_CATEGORY; 
            this.numLayers = numLayers;
            this.lineDividerRenderer = lineDividerRenderer;
            initialize();            
        }
        
        public void focusOnFilter (Predicate inFilter,  int inLeftBorder, int inWidth, String inAnimateAction ){
        	categoryKey = CytoPrefuse.DISPLAY_LAYER;
        	filter = inFilter;
           width = inWidth;
           leftBorder = inLeftBorder;
           animateAction = inAnimateAction;
           initialize();
           firstRun = false;
        }
        
        public boolean pointOccupied(int x, int y){
        	return nodeAtPosition[x/SCALE][y/SCALE] != null;
        }
        public NodeItem getNodeAtPosition (int x, int y) {
         	return (NodeItem)nodeAtPosition[x/SCALE][y/SCALE];
            
        }
        private void checkNodeAtPos() {
        	Set nodesSeen = new HashSet();
        	for (int x = 0; x < nodeAtPosition.length; x++){
        		for (int y = 0; y <nodeAtPosition[0].length; y++){
        			Node n = nodeAtPosition[x][y];
        			if (n!=null){
        				if (nodesSeen.contains(n)){
        					System.out.println("NODESSEEN inconsisten");
        				}else {
        					nodesSeen.add(n);
        				}
        			}
        		}
        	}

        	Set uniqueNodeLoc = new HashSet();
            
        	for (int i =0; i < nodepos.length; i++){
        		Point p = nodepos[i];
	       		if (p != null){	        		
	        		Integer hash = new Integer (p.x*10000+p.y);
		        	if (hash.intValue() != 0){
		        		if (uniqueNodeLoc.contains(hash)){
		        			System.out.println("nodepos contains two different points that map to the same location");
		        			
		        		}
		        		uniqueNodeLoc.add(hash);
		        	}
	        	}
	        }
        }
        
        private void initialize() {
        	firstRun = true;
        	for (int i =0; i < nodepos.length; i++){
        		nodepos[i] = new Point();
        	}
        	 totalImprovements =0;
            iteration = 1;
            iterationTime = 0;
            iterationsWithoutImprovement = 0;
            //add 1 for  the downstream layer
            nodeCounts = new int [numLayers + 1];
            Arrays.fill(nodeCounts, 0);
              
            totalNodesFiltered =0;
            TupleSet nodes = graph.getNodes();
            
            int anyLayerCount =0;
            Iterator nodesIt = nodes.tuples();
            while (nodesIt.hasNext()){
                NodeItem node = (NodeItem) nodesIt.next();

                if (node.getBoolean(CytoPrefuse.ANY_LAYER) && ! node.getBoolean(CytoPrefuse.DOWNSTREAM_GENE)){
                	anyLayerCount++;
                } else{
                	if (node.getBoolean(CytoPrefuse.DOWNSTREAM_GENE)){
                		  nodeCounts[numLayers]++;
	                }else{
		                int bioCategory = node.getInt(CytoPrefuse.BIO_CATEGORY);
		                nodeCounts[bioCategory]++;
	                }
                }
            }
            

             nodesIt = nodes.tuples(filter);
             while (nodesIt.hasNext()){
                 totalNodesFiltered++;
                 nodesIt.next();
             }

           //  System.out.println(totalNodesFiltered + " pass the filter");
                     
            nodesToSample = new ArrayList();                        
            nodesIt = nodes.tuples(filter);
            while (nodesIt.hasNext()){
                NodeItem node = (NodeItem) nodesIt.next();
                nodesToSample.add(node);
            }
        
                 
            nodeSampling = new ArrayList ((int) Math.sqrt(totalNodesFiltered));
            pointSampling = new ArrayList ((int)Math.sqrt(totalNodesFiltered));
            
            gridHeight = (int) (2 * Math.sqrt( nodes.getTupleCount()));            
            int gridWidth = gridHeight;
                                         
            layerBounds = new int [numLayers +1];
            int bottom = 0;
            int top = 0;
              
            float avgAnyLayersPerLayer = anyLayerCount / (numLayers+1);
            for (int i = 0; i < numLayers + 1; i++){
                bottom = top;                           
                       
                top += (int) (((nodeCounts[i]+ avgAnyLayersPerLayer )/(float)nodes.getTupleCount()) * gridWidth) ; 
                if (top - bottom == 0){
                    top++;
                    gridHeight++;
                }
                if (top - bottom == 1) {
                    top++;
                    gridHeight++;
                }
                if (top - bottom == 2) {
                    top++;
                    gridHeight++;
                }
                layerBounds[i] = top;
            }
            //System.out.println ("Using a grid size of [" +gridWidth + ":" + gridHeight + "]");
                    
            layerBounds[layerBounds.length -1 ] =gridHeight;
            lineDividerRenderer.setDividers(layerBounds); 
            
            int rightBorder = leftBorder + width;                   
               
            candidatePoints = new ArrayList [numLayers +1];               
            for (int i=0; i <numLayers +1; i++){
                candidatePoints[i] = new ArrayList();
            }

            bottom = 0;
            top = 0;
            for (int i = 0; i < nodeCounts.length; i++){
                bottom = top;
                top = layerBounds[i];
                for (int j=leftBorder+1; j < rightBorder; j++){
                    for (int k=bottom+1; k < top; k++){
                        Point p = new Point (j,k);
                        candidatePoints[i].add( (p));
                    }
                }
            }    
       
            nodesIt = nodes.tuples();	            
            int maxGridPoint = Integer.MIN_VALUE;
            while (nodesIt.hasNext()){
                NodeItem node = (NodeItem) nodesIt.next();      	                
                int x = (int) node.getEndX();
                if (x/SCALE > maxGridPoint){
                	maxGridPoint = x/SCALE;
                }
                int nx =nodepos[node.getRow()].x;
                if (nx > maxGridPoint){
                	maxGridPoint = nx;
                }
            }
            maxGridPoint = (maxGridPoint > gridHeight)? maxGridPoint : gridHeight;
            
            initializeNodeAtPosition (maxGridPoint , gridHeight);
            assignNodePositions (nodes, leftBorder, rightBorder, candidatePoints);     
            initializeSpaceGrid();

			 checkNodeAtPos();
		}

        public int[] getLayerBounds() {
        	return layerBounds;
        }
        
        public void run(double frac) {

        	if (firstRun) {
        		zoomBounds.setRect( findFilterBounds());
        		DisplayLib.fitViewToBounds(m_vis.getDisplay(0), zoomBounds, 1000);
        		firstRun = false;
        	}
        	TupleSet nodes = graph.getNodes();
            for (int i=0; i < 10; i++)
                improveLayout (nodes, candidatePoints); 
            
            if (iteration%100 ==0){
                CytoPrefuse cytoPrefuse = (CytoPrefuse) m_vis.getDisplay(0);                                 
                EdgeBundler edgeBundler = cytoPrefuse.edgeBundler;
                if (edgeBundler != null){
               		edgeBundler.recomputeLocations();
               		m_vis.run("color");
            	}
                m_vis.run("aggLayout");                
            	
            }
      
            if (iterationsWithoutImprovement > totalNodesFiltered) {
            
                //We're done.
                m_vis.cancel(cancelAction);
                m_vis.run("layoutPostProcess");

               
            }
        }
        
	
		public Rectangle2D findFilterBounds() {
			int minY = Integer.MAX_VALUE;
			int maxY = Integer.MIN_VALUE;
			Iterator iter = graph.getNodes().tuples(filter);
		     while ( iter.hasNext() ) {
		         NodeItem node = (NodeItem)iter.next();
		         Point np = nodepos[node.getRow()];
		         if (np.y > maxY){
		        	 maxY = np.y;
		         }
		         if (np.y  < minY){
		        	 minY = np.y;
		         }  	
		     }
		     Rectangle2D bounds = new Rectangle2D.Float((leftBorder -  2) * SCALE,  (minY - 2) *SCALE, (width+4) * SCALE, (maxY -minY+ 4) *SCALE);
		     return bounds;
		}

		private void initializeSpaceGrid() {
		
		    //the spacegrid needs to be big enough for any node to be tracked,
		    //not just the nodes we're laying out (for the nodes we are laying
		    //out could join to nodes outside the region.)
		    int maxX = Integer.MIN_VALUE;
		    int maxY = Integer.MIN_VALUE;
		    TupleSet nodes = graph.getNodes( );
		    Iterator nodesIt = nodes.tuples();
		    while (nodesIt.hasNext()){
		        NodeItem node = (NodeItem)nodesIt.next();
		        Point p = nodepos[node.getRow()];
		        if (p.x > maxX){
		            maxX = p.x;
		        }
		        if (p.y > maxY){
		            maxY = p.y;
		        }
		    }               
		    gridWidth = maxX < leftBorder + width ? leftBorder+ width : maxX;
		    gridHeight = maxY < gridHeight ? gridHeight : maxY;
		    spaceGrid = new SpaceGrid (gridWidth + 1, gridHeight + 1, graph.getNodeCount(), this);
		        
		        
		    Iterator iter = graph.getNodes().tuples(filter);
		    Set edgesAdded = new HashSet();
		    while ( iter.hasNext() ) {
		        NodeItem node = (NodeItem)iter.next();
		        spaceGrid.addNode(node);
		        Iterator edgesIt = node.edges();
		        while (edgesIt.hasNext()){
		            Edge edge = (Edge)edgesIt.next();
		            if (!edgesAdded.contains(edge)){
		                spaceGrid.addEdge(edge);
		                edgesAdded.add(edge);
		            }
		        }
		    }
		 }
        
        private void initializeNodeAtPosition(int x, int y) {
            nodeAtPosition = new Node[x+1][y+1];
            for (int xi=0; xi <= x; xi++) {
                Arrays.fill(nodeAtPosition[xi],null);
            }
        }

        public final Point getPositionOfNode (Node n){
        	return nodepos[n.getRow()];
        }
        

        private void improveLayout(TupleSet nodes, List[] points) {
                
            long startTime = System.currentTimeMillis();
            int bestImprovement =0;
            NodeItem bestImprovedNode = null;
            Point bestNewPoint = null;
            iteration ++;
            
            bestImprovement = 0;
            //let's take a random Root(n) sampling of the nodes
            nodeSampling.clear();
            for (int i=0; i < Math.sqrt(nodesToSample.size()); i++) {
                int sample = generator.nextInt(nodesToSample.size());
                nodeSampling.add(i, nodesToSample.get( sample));                                
            }

            for (int i=0; i < nodeSampling.size(); i++) {
                NodeItem node = (NodeItem) nodeSampling.get(i);
                removeNodeFromSpaceGrid(node);
                int oldScore = scoreNode (node);
                        
                //And take a random sampling of the possible points
                pointSampling.clear();
                int sampleSize =0;                              
              
                int layer =0;
                if (node.getBoolean(CytoPrefuse.DOWNSTREAM_GENE)){
                	layer = numLayers;
                }else {
                	layer =  node.getInt(categoryKey);                    
                }
                boolean anyLayer =node.getBoolean(CytoPrefuse.ANY_LAYER) && !node.getBoolean(CytoPrefuse.DOWNSTREAM_GENE); 
                if (anyLayer){
                    sampleSize = (int) Math.sqrt(nodesToSample.size());                      	
                }else {
                   sampleSize = (int) Math.sqrt(points[layer].size());                     	
                }
               
                for (int  j=0; j< sampleSize; j++){                	
                	  if (anyLayer){
                        layer = generator.nextInt(numLayers);
                    }
                    int sample = generator.nextInt(points[layer].size());
                    pointSampling.add(j, points[layer].get(sample));
                }
                
                                
                Iterator pointIt = pointSampling.iterator();
                nodeOrigSave.x = nodepos[node.getRow() ].x;
                nodeOrigSave.y = nodepos[node.getRow() ].y;
                
                int improvement = 0;
                while (pointIt.hasNext()){      
                        
                    Point trialPoint = (Point)pointIt.next();
                    int newScore;
                    if (pointIsEmpty(trialPoint)) {
                        //Great!  nothing there.  try it out.                           
                        moveNodeToEmptyPoint(node, trialPoint, false);
                        improvement = oldScore - scoreNode (node);
                        moveNodeToEmptyPoint(node, nodeOrigSave, false);                                               
                    } else{
                    	
                        Node swapNode = nodeAtPosition[trialPoint.x][trialPoint.y];             
                        if (swapNode.getRow() == node.getRow()) {
                            improvement = 0;
                        }else{
                        	if (canSwapNodes (swapNode, node)){                            	
                                   removeNodeFromSpaceGrid(swapNode);
	                                    swapNodePositions(swapNode, node, false);       
	                                addNodeToSpaceGrid(swapNode);
	                                newScore = scoreNode (node);
	                        
	                                int nodeImprovement = oldScore - newScore;
	                                                        
	                                removeNodeFromSpaceGrid(swapNode);
	                                    addNodeToSpaceGrid(node);
	                                        int swapNewScore = scoreNode ((NodeItem) swapNode);
	                                                        
	                                        removeNodeFromSpaceGrid(node);
	                                            swapNodePositions(swapNode, node, false);
	                                        addNodeToSpaceGrid(node);
	                                        int swapOldScore = scoreNode((NodeItem) swapNode);                                              
	                                    removeNodeFromSpaceGrid(node);
	                                addNodeToSpaceGrid(swapNode);
	                                
	                                int swapImprovement = swapOldScore - swapNewScore;
	                                improvement = nodeImprovement + swapImprovement;
                            	   
                             }else {//different categories.  Can't swap 
                                improvement = 0;                                                
                            }   
                            
                            
                        		//	improvement  = 0;    
                        }//end else (different nodes.  consider swap)
                        
                   
                    }//end point isn't empty.  Try swapping
                        
                    if (improvement > bestImprovement){
                        bestImprovedNode = node;
                        bestNewPoint = trialPoint;
                        bestImprovement = improvement;
                    }
                } //while end.  Try next trial point  
                addNodeToSpaceGrid(node);
                               
                if (bestImprovement > 0){              
                
                     iterationsWithoutImprovement =0;
                    totalImprovements += bestImprovement;
            
                    if (pointIsEmpty(bestNewPoint)) {
                         moveNodeToEmptyPoint(bestImprovedNode, bestNewPoint, true);
                         setDisplayFromNodepos(bestImprovedNode);
                    }else  {   //point is occupied                    
                        NodeItem swapNode = (NodeItem) nodeAtPosition[bestNewPoint.x][bestNewPoint.y];                  
                        swapNodePositions(bestImprovedNode, swapNode, true);
                        setDisplayFromNodepos(bestImprovedNode);
                        setDisplayFromNodepos(swapNode);                                                                       	   
                    }//end swap conditions
                    	
                	bestImprovement =0;
                    
                } else {//best improvement is not greater than 0
                    iterationsWithoutImprovement++;
                }
               // spaceGrid.checkHealth();
            }//end for.  Sample next node
        
                                

            long itTime = System.currentTimeMillis() - startTime;
            iterationTime += itTime;
            
            								
            //cytoPrefuse.statusBar.setStatusMessage();
            
            if (iteration % 50==0){
            	/* 
                 long estimatedTimeLeft = (totalNodesFiltered * 6
                 							*	(iterationTime/iteration)
                 							- iterationTime) 
                 							/ 1000;
                 String message = "Computing optimal layout.  Estimated time left ";
                 if (estimatedTimeLeft > 60) {
                	 message += estimatedTimeLeft / 60 + " minutes";
                 }else {
                	 message += " < " + ((estimatedTimeLeft  /15)+1) * 15 + " seconds";
                 }
                 
                 CytoPrefuse cytoPrefuse = (CytoPrefuse) m_vis.getDisplay(0);                 
                 cytoPrefuse.statusBar.setStatusMessage(message);
                 */
            	System.out.println ( iteration + ": Iterations without improvement " + iterationsWithoutImprovement + ": Current iteration " + itTime + "ms.  Average iteration " + iterationTime/iteration);                                   
            }                       
        }
                
        private boolean canSwapNodes(Node a, Node b) {
        	boolean aDownstream = a.getBoolean(CytoPrefuse.DOWNSTREAM_GENE);
        	boolean bDownstream = b.getBoolean(CytoPrefuse.DOWNSTREAM_GENE);
        	//if only one of the  nodes is a downstream node...
        	if (aDownstream ^ bDownstream)
        		return false;
        	
        	boolean aAnyLayer = a.getBoolean(CytoPrefuse.ANY_LAYER);
        	boolean bAnyLayer = b.getBoolean(CytoPrefuse.ANY_LAYER);
        	
        	if (aAnyLayer && bAnyLayer)
        		return true;
          	//if only one of the nodes is allowed to move to any layer            
        	if (aAnyLayer ^ bAnyLayer)
        		return false;
        	
        	//they're not both anylayer, and they're not both downstream
        	//genes, so they better have the same bio_category
        	int aCat = a.getInt(categoryKey);
        	int bCat = a.getInt(categoryKey);
        	return aCat==bCat;
        	
        	
		}

		public int getProgress() {
        	return (int) ((iteration / (totalNodesFiltered*6.0f)) *1000f);        	
        }
        public String getProgressMessage(){
        	  long estimatedTimeLeft = (totalNodesFiltered * 6
						*	(iterationTime/iteration)
						- iterationTime) 
						/ 1000;
			String message = "Computing optimal layout.  Estimated time left ";
			if (estimatedTimeLeft > 60) {
				message += estimatedTimeLeft / 60 + " minutes.";
			}else {
				message += " < " + ((estimatedTimeLeft  /15)+1) * 15 + " seconds.";
			}
			return message;        	
        }
        

        private void addNodeToSpaceGrid(Node node) {
            spaceGrid.addNode((NodeItem) node);             

            Iterator edges = node.edges();
            while (edges.hasNext()) {
                Edge e = (Edge) edges.next();
                spaceGrid.addEdge(e);
            }               
        }
        

               
        private void removeNodeFromSpaceGrid(Node node) {
            Iterator edges = node.edges();
            while (edges.hasNext()) {
                    Edge e = (Edge) edges.next();
                    spaceGrid.deleteEdge(e);
            }
            spaceGrid.deleteNode((NodeItem) node);
        }

		
        private int scoreNode(NodeItem node) {
            int score = 0;
            if ( node.getBoolean(CytoPrefuse.DOWNSTREAM_GENE)) {                        
                score = countCrossings(node, true);          
                score += 5 * gridHeight* countNeighbourNodeDistance (node);
                score += 500* gridHeight * countFunctionNeighbourNodeDistance (node);
                score += 500*gridHeight * countOtherFunctionInNeighbourhood(node);
            }else{
                score = countCrossings(node, false);                     
                score += 5 * gridHeight * countNeighbourNodeDistance (node);                              
                score += 5*gridHeight* countFunctionNeighbourNodeDistance(node);   
                //score += gridHeight * countOtherFunctionInNeighbourhood(node);
            }               
            return score;
        }

		private int countNeighbourNodeDistance (NodeItem node, Iterator neighbs) {
            int score =0;
            Point nodePos = nodepos[node.getRow()];
            while (neighbs.hasNext()){
                Node neighbour = (Node) neighbs.next();
                boolean checkNeighbour = true;
                if (filter != null){
                	checkNeighbour = filter.getBoolean(neighbour);
                }
                if (checkNeighbour){
	                Point neighbPos = nodepos[neighbour.getRow()];
	                int increment =Math.abs ( neighbPos.x - nodePos.x)  +Math.abs( neighbPos.y - nodePos.y);        
	                score += increment;
                }
            }
            return score;
        }

        private int countFunctionNeighbourNodeDistance(NodeItem node) {
            List functionNeighbourList = (List) node.get(CytoPrefuse.FUNCTION_NEIGHBOURS);
            int score =0;
            if (functionNeighbourList != null) {
            	score = countNeighbourNodeDistance(node, functionNeighbourList.iterator());           
            }
            return score;
        }

        private int countOtherFunctionInNeighbourhood(NodeItem node) {
            int score = 0;
            boolean functionUnknown = node.getBoolean(CytoPrefuse.BIO_FUNCTION_UNKNOWN);
            int bioFunction = node.getInt(CytoPrefuse.BIO_FUNCTION);
            Point pos = nodepos[node.getRow()];
            for (int x = -1; x <=1; x++) {
                for (int y =-1; y <=1; y++) {
                    neighbPoint.x = pos.x + x;
                    neighbPoint.y = pos.y + y;
                    if ( neighbPoint.x>= 0 && neighbPoint.x < nodeAtPosition.length &&
                         neighbPoint.y >= 0 && neighbPoint.y < nodeAtPosition[0].length &&
                         (x!=0 || y!=0) ){
                        Node neighbour = nodeAtPosition[neighbPoint.x][neighbPoint.y];
                        if (neighbour != null) {
                            int neighbBioFunction = neighbour.getInt(CytoPrefuse.BIO_FUNCTION);
                            boolean neighbFunctionUnknown = neighbour.getBoolean(CytoPrefuse.BIO_FUNCTION_UNKNOWN);
                            if (!functionUnknown && !neighbFunctionUnknown){
                            	if (neighbBioFunction != bioFunction ) {
                            		score ++;
                            	}
                            }else if (functionUnknown ^ neighbFunctionUnknown){
                            	score++;
                            }
                        }
                    }
                }
            }
            return score;

        }

        private int countNeighbourNodeDistance(NodeItem node) {
                return countNeighbourNodeDistance(node, node.neighbors());
        }

        private int getLayer(int yCoord) {
            int testY = yCoord;
            for (int i =0; i < layerBounds.length; i++){
                if (testY < layerBounds[i])
                    return i;
            }
            return -1;
        }

        private int countCrossings(NodeItem node, boolean ignoreEdgeEdgeCrossings) {
        
            int edgeCrossings = 0;
            Iterator edgesIt = node.edges();
            while (edgesIt.hasNext()){      
                Edge a = (Edge) edgesIt.next();
                if (ignoreEdgeEdgeCrossings) {
                    edgeCrossings += spaceGrid.countCrossings (a, 10, 500 * gridHeight);                            
                }else {
                    edgeCrossings += spaceGrid.countCrossings (a, 40 * gridHeight, 500 * gridHeight);
                }
            }               
            
            int thisNodeEdgeCrossings = spaceGrid.countCrossings(nodepos[node.getRow()], 500*gridHeight);
            return edgeCrossings + thisNodeEdgeCrossings;
        }
        
        private void assignNodePositions(TupleSet nodes, int leftBorder, int rightBorder, List[] points) {
            Random generator = new Random();
                    
            Set unplaceableNodes = new HashSet();
          
            Iterator nodesIt = nodes.tuples(new NotPredicate(filter));		            
            //Place the nodes to their closest gridpoint
            while (nodesIt.hasNext()){
                NodeItem node = (NodeItem) nodesIt.next();      
                
                boolean couldSetPosition = setNodeposFromDisplay(node, Integer.MIN_VALUE, Integer.MAX_VALUE);
                if (!couldSetPosition){
                    unplaceableNodes.add(node);
                }                       
            }
            //And place the filtered nodes within the boundaries
            nodesIt = nodes.tuples(filter);
            
            //Place the nodes to their closest gridpoint
            while (nodesIt.hasNext()){
                NodeItem node = (NodeItem) nodesIt.next();      
                
                boolean couldSetPosition = setNodeposFromDisplay(node, leftBorder, rightBorder);
                if (!couldSetPosition){
                    unplaceableNodes.add(node);
                }
            }
            
            //nodes that weren't previously on the grid get placed at random
            
            nodesIt = unplaceableNodes.iterator();
            while (nodesIt.hasNext()){
                NodeItem node = (NodeItem) nodesIt.next();                      
                int layer = getLayer(generator, node);
                int randPos;
                Point pCandidate;
                do{
                    randPos = generator.nextInt( points[layer].size());             
                    pCandidate = (Point) points[layer].get(randPos);
                }while (!pointIsEmpty(pCandidate));
                
                Point p = new Point (pCandidate.x, pCandidate.y);
                nodepos[node.getRow()]=p;
	            nodeAtPosition[p.x][p.y] = node;
                setDisplayFromNodepos(node);
            }
	     }

        private void setDisplayFromNodepos(NodeItem node){
        	Point p = nodepos[node.getRow()];
        	 if (animateAction == null){
        		 node.setX(p.x*SCALE);
        		 node.setY(p.y *SCALE);
        	 }
             node.setEndX(p.x *SCALE);
             node.setEndY(p.y *SCALE);                           
             node.setInt(CytoPrefuse.DISPLAY_LAYER, getLayer(p.y)+1);
        }
		private boolean setNodeposFromDisplay(NodeItem node, int left, int right) {
			int x = (int) node.getEndX();
			int y = (int) node.getEndY();
           
			Point p = nodepos[node.getRow()];
			p.x = x/SCALE;
			p.y = y/SCALE;
			if (p.x < 0 || p.x > nodeAtPosition.length-1 ||
					p.y < 0 || p.y > nodeAtPosition[0].length -1 ){
				return false;
			}
			boolean couldSetPosition = false;
			if (x%SCALE ==0 && y%SCALE == 0){
				if (p.x > left && p.x < right){
				    if (pointIsEmpty(p)){
				        //see if it's a legal layer for the node
				        int layer = getLayer (p.y);
				        boolean legal = false;
				        if (node.getBoolean(CytoPrefuse.DOWNSTREAM_GENE)){
				        	legal = (layer == numLayers);
				        }else {					        	
					        int bioCategory = node.getInt(CytoPrefuse.BIO_CATEGORY);
					        legal =node.getBoolean(CytoPrefuse.ANY_LAYER) ||
					        	   bioCategory == (layer);
				        }
				        if ( legal ) {
				            couldSetPosition =true;
				            nodeAtPosition[p.x][p.y] = node;            
				            node.setInt(CytoPrefuse.DISPLAY_LAYER, layer+1);
				        }else {
				        	//	System.out.println("Illegal layer");
				        }
				    }else {
	//			    	System.out.println("Point wasn't empty");
				    }			
				}else {
		//			System.out.println("Outside bounds");
				}
			}
			return couldSetPosition;
		}

        private void moveNodeToEmptyPoint(Node node, Point p, boolean updateGrid) {
            if (updateGrid)
                removeNodeFromSpaceGrid(node);
            
            Point oldPos =  nodepos[node.getRow()];
            nodeAtPosition[oldPos.x][oldPos.y] = null;
                
            oldPos.x = p.x;
            oldPos.y = p.y;
                
            nodeAtPosition[p.x][p.y] = node;        
            if (updateGrid)
                addNodeToSpaceGrid(node);
        }
        
        private void swapNodePositions (Node n1, Node n2, boolean updateGrid) {
                
            Point n1Pos = nodepos[n1.getRow()];
            Point n2Pos = nodepos[n2.getRow()];
           
            if (updateGrid){
                removeNodeFromSpaceGrid(n1);     
               removeNodeFromSpaceGrid(n2);                
            }
           
            //update the internal map
            nodeAtPosition[n2Pos.x][n2Pos.y] = n1;
            nodeAtPosition[n1Pos.x][n1Pos.y] = n2;
         
            int x = n1Pos.x;
            int y = n1Pos.y;
    
            n1Pos.x = n2Pos.x;
            n1Pos.y = n2Pos.y;
                     
            n2Pos.x = x;
            n2Pos.y = y;
                  
            if (updateGrid){
                addNodeToSpaceGrid(n2);
                addNodeToSpaceGrid(n1);                  	                
            }
        }
        



		private boolean pointIsEmpty(Point p) {
			if (nodeAtPosition[p.x][p.y] != null)
				return false;
			else{
				//make sure it's not a layer boundary
				for (int i=0; i < layerBounds.length; i++) {
					if (p.y == layerBounds[i])
						return false;					
				}
				return true;
			}
        }

        private int getLayer(Random generator, Node node) {
            int layer =0;
            if (node.getBoolean(CytoPrefuse.ANY_LAYER) && !node.getBoolean(CytoPrefuse.DOWNSTREAM_GENE)){
                    int yPoint= generator.nextInt (gridHeight);
                    layer = getLayer(yPoint);
            }else {
            	if (node.getBoolean(CytoPrefuse.DOWNSTREAM_GENE))
            		layer= numLayers;
            	else {
            		layer = node.getInt(categoryKey);
            	}
            }
            return layer;
        }

		public int getGridWidth() {
			return gridWidth;
		}
		
		
		public int getGridHeight() {
			return gridHeight;
		}

		public Point pointLocation(VisualItem item) {
			return nodepos[item.getRow()];
		}

		public void setNodeAtLocation(VisualItem item, int unscaledx, int unscaledy) {
			Point scaledPoint = new Point(unscaledx/SCALE, unscaledy/SCALE);
			moveNodeToEmptyPoint((Node) item, scaledPoint, false);
			item.setInt(CytoPrefuse.DISPLAY_LAYER, getLayer(scaledPoint.y)+1);
		}

     
}
